#!/usr/bin/python

import re
import os
import sys
import string

import asmcallbacks

from asmconsts import *
from asmclasses import *

PROGRAM_NAME = "Decomp"
VERSION = "0.1"

def generateObjdumpFile(bin, out):
    os.putenv("LANG", "C")
    disassembled_data = os.popen("objdump --prefix-addresses -d %s" % bin).readlines()

    return disassembled_data

def readRodataSection(bin):
    os.putenv("LANG", "C")
    disassembled_data = os.popen("objdump -s %s" % bin).readlines()

    buf = []
    bStart = False

    for line in disassembled_data:
        line = line.replace("\r", "").replace("\n", "")
        
        if line.lower().find("contents of section .rodata") == 0 and bStart == False:
            bStart = True
        else:
            if bStart and line.lower().find("contents of section") == 0:
                break
            elif bStart:
                buf.append(line)

    return rodata2hex(buf)

def rodata2hex(lines):
    
    i = -1
    address = ""
    data = ""

    for line in lines:
        i += 1

        line = line.strip()
        line = line[:line.find("  ")]
        
        res = line.split(" ")

        for x in res:
            if x == res[0]:

                if address == "":
                    x = eval("0x" + x)
                    address = x
            else:
                data += x

    data = data.replace(" ", "")
    objRodata = Rodata(address, str2hex(data))

    return objRodata

def str2hex(arg):

    i = 0
    buf = ""

    while 1:
        buf += eval("'\\x" + arg[i:i+2] + "'")
        i += 2
        
        if i == len(arg):
            break

    return buf

def checkBinFile(bin):
    try:
        f = file(bin, "r")
        f.close()
    except:
        print "*** Error opening binary file %s" % bin
        print sys.exc_info()[1]
        sys.exit(0)

def splitData(asm):
    
    address = asm[0:8]

    pos = asm.find("<")
    end = asm.find(">")
    function = asm[pos+1:end]
    pos = function.find("+")
    functionPos = function

    if pos == -1:
        pos = function.find("-")

    if pos > -1:
        function = function[:pos]

    code = asm[end+2:]

    return CDisassembledLine(address, function, code, functionPos)

def isProbablyModifiable(instruction, current):
    """
    Check if the first parameter of the function appears to be influenctiable (i.e., malloc, free).
    """
    current = current.replace("\t", "").replace(" ", "").lower()
    instruction = instruction.lower().replace("\t", "").replace(" ", "")

    for y in DOUBLE_CHECK_CALLS:
        if current.find(y) > -1:
            for x in USER_CONTROLLED_DATA[CURRENT_CPU]:
                if instruction.find(x) > -1:
                    return y
    
    return False

def searchVar(code, vars):

    x = re.compile("(0x[a-f0-9]{1,8})", re.IGNORECASE)
    res = x.findall(code)
    
    if res:
        address = eval(res[0])
    else:
        return

    if address > vars.address and address < len(vars.data)+vars.address:
        buf = ""
        index = address-vars.address

        buf = vars.data[index::-1]
        index -= buf.find("\x00")-1

        buf = ""

        for c in vars.data[index:]:
            if c != "\x00":
                buf += c
            else:
                return buf
    else:
        return address

def appearsToBeVariable(disLine):
    if not disLine.code.find("$") > -1:
        return False
    
    if disLine.code.find("0x") > disLine.code.find("$"):
        if disLine.code.find(",") >= disLine.code.find("0x")+2:
            return True
        elif disLine.code.find("(") >= disLine.code.find("0x")+2:
            return True
        else:
            for c in disLine.code[disLine.code.find("0x")+2:]:
                if c not in string.hexdigits and c not in ["\r", "\n", " "]:
                    return False

            return True

    return False

def parseLines(data, showAddress = False, showReport = False, vars = None, showStrings = False, pseudoC = False):

    repbuf = "Report generated from " + PROGRAM_NAME + " " + VERSION
    repbuf += "\r\n" + "-"*len(repbuf) + "\r\n\r\n"
    hits = 0

    buf = "; Source generated from " + PROGRAM_NAME + " " + VERSION + "\r\n"
    bContinue = True
    oldFunction = ""

    """
    Check all the interesting addresses for vulnerability development
    
        call *%(eax, edx, ebx, etc...), jmp, etc...

    """
    buf += "\r\n"
    buf += "; Interesting addresses\r\n;\r\n;"

    for line in data:
        line = line.strip("\r").strip("\n")
        
        disLine = splitData(line)
        
        for x in INTERESTING_INSTRUCTIONS[CURRENT_CPU]:
            if disLine.code.replace(" ", "").replace("\t", "").find(x) > -1:
                buf += "\t" + disLine.code + "; 0x" + disLine.address + " <" + disLine.functionPos + "> "
                
                if disLine.address.find("00") > -1:
                    buf += "NOTE: May be unusable???"
                
                buf += "\r\n;"

    buf += "\r\n"

    """
    Show ONLY the functions the programmers wrote, or at least try doing it ;)
    """
    
    # Last asm block delimited by jmp, call or ret (x86)
    lastBlock = ""
    lastBlockLines = []
    # Previous asm instruction
    previousInstruction = ""
    i = -1
    for line in data:
        i += 1
        line = line.strip("\r").strip("\n")

        if bContinue:
            # Ok, we found the .text section
            if line.find(" .text:") > -1:
                bContinue = False
                continue
            else:
                continue
        else:
            if line.find("section .") > -1 and line.find(":") > line.find("section ."):
                break

        if line == "":
            continue

        disLine = splitData(line)

        if disLine.code.replace(" ", "").replace("\t", "") in IGNORE_INSTRUCTIONS[CURRENT_CPU]:
            continue

        bIgnore = disLine.function in IGNORE_FUNCTIONS
        varCode = False

        if disLine.function != oldFunction:

            if not bIgnore:
                if oldFunction != "" and oldFunction not in IGNORE_FUNCTIONS:
                    buf += "END ASM;\r\n"

                if showAddress:
                    buf += "\r\nFUNCTION %s AT ADDRESS 0x%s\r\nBEGIN ASM\r\n" % (disLine.function, disLine.address)
                else:
                    buf += "\r\nFUNCTION %s\r\nBEGIN ASM\r\n" % disLine.function

                if showAddress:
                    buf += "\t0x%s <%s>: %s"  % (disLine.address, disLine.functionPos, disLine.code)
                else:
                    buf += "\t%s"  % disLine.code

            oldFunction = disLine.function

        else:
            if not bIgnore:
                if showAddress:
                    buf += "\t0x%s <%s>: %s" % (disLine.address, disLine.functionPos, disLine.code)
                else:
                    buf += "\t%s" % disLine.code

        if not bIgnore:
            lastBlock += "\t0x%s <%s>:\t%s\r\n" % ((disLine.address, disLine.functionPos, disLine.code))
            objAsm = AsmLine()
            objAsm.address = disLine.address
            objAsm.code = disLine.code

            if CURRENT_CPU == "x86":
                if appearsToBeVariable(disLine):
                    varCode = searchVar(disLine.code, vars)
                    
                    if x:
                        buf += "\t\t; %s " % repr(varCode)
                        lastBlock = lastBlock[0:len(lastBlock)-2] + "\t\t; %s\r\n" % repr(varCode)
                        objAsm.data = varCode

            lastBlockLines.append(objAsm)

            res = asmcallbacks.codeCallback(lastBlockLines, disLine.code, x, previousInstruction)
            if res:
                buf += "\t\t; CODE: %s" % res

            for x in WARNING_CALLS:
                if disLine.code.find(x) > -1 and disLine.function == oldFunction:
                    hits += 1
                    if showReport:
                        repbuf += "\r\n%d) Usage of %s in function %s:\r\n\r\n" % (hits, x, disLine.function)
                        repbuf += lastBlock
                        repbuf += "\r\n\r\n"

                        res = asmcallbacks.warningCallback(lastBlockLines, disLine.code, x, previousInstruction)

                        if res:
                            repbuf += "Analisis: %s\r\n\r\n" % res

                    buf += "\t\t; CHECK: Use of %s (Hit %d)" % (x, hits)

            func = isProbablyModifiable(previousInstruction, disLine.code)
            if func != False:
                buf += "\t\t; CHECK: Use of %s with user supplied data?" % func
                
                if showReport:
                    hits += 1
                    repbuf += "%d) Warning. Usage of %s in function %s with (probably) user supplied data.\r\n" % (hits, func, disLine.functionPos)

        if not bIgnore and oldFunction not in IGNORE_FUNCTIONS:
            buf += "\r\n"

        if (disLine.code[0] == "j" or disLine.code.find("call") > -1 or disLine.code.find("ret") > -1) and not bIgnore:
            buf += "\r\n"
            lastBlock = ""
            lastBlockLines = []

        previousInstruction = disLine.code

    buf += "END ASM;\r\n"
    repbuf += "\r\nTotal of %d hits. Happy haunting!\r\n" % hits

    return buf, repbuf

def decompile(bin, out, showAddress = False, showReport = False, showStrings = False, pseudoC = False):
    checkBinFile(bin)
    data = generateObjdumpFile(bin, out)
    vars = readRodataSection(bin)

    output, report = parseLines(data, showAddress, showReport, vars, pseudoC)

    if out is None:
        print output

        if showReport:
            print
            print report
        
        if showStrings:
            os.system("strings %s" % bin)

    else:
        try:
            f = file(out + ".gen", "w")
            f.write(output)
            f.close()

            if showReport:
                f = file(out + ".rep", "w")
                f.write(report)
                f.close()
            
            if showStrings:
                os.system("strings %s > %s.str" % (bin, out))

        except:
            print "*** Error writing to file"
            print sys.exc_info()[1]

    return output, report

def usage():
    print PROGRAM_NAME, "version", VERSION
    print "Copyright (c) 2007 Joxean Koret"
    print
    print "Usage:"
    print sys.argv[0], "<binary file> [<output file>] [options]"
    print
    print "Options:"
    print "\t-a\tShow address"
    print "\t-r\tGenerate a report (.rep) of the most interesting functions"
    print "\t-s\tGenerate an string file (.str)"
    print

def main():

    if len(sys.argv) < 2:
        usage()
        sys.exit(0)

    i = 0

    binaryFile = None
    outputFile = None
    showAddress = False
    showReport = False
    showStrings = False
    pseudoC = True

    for arg in sys.argv[1:]:
        i += 1

        if i == 1:
            binaryFile = arg
        elif i == 2 and arg[0] != "-":
            outputFile = arg
        else:
            if arg.lower() in ["--help", "-h", "--h"]:
                usage()
                sys.exit(0)
            elif arg.lower() in ["-a", "--address"]:
                showAddress = True
            elif arg.lower() in ["--report", "-r"]:
                showReport = True
            elif arg.lower() in ["-s", "--strings"]:
                showStrings = True
            elif arg.lower() in ["-p", "--pseudo-c"]:
                pseudoC = True
            else:
                print "*** Unknown option '%s'" % arg
                print
                usage()
                sys.exit(0)

    if binaryFile is None:
        print "*** Bad command line"
        print
        usage()
        sys.exit(0)
    else:
        decompile(binaryFile, outputFile, showAddress, showReport, showStrings, pseudoC)

if __name__ == "__main__":
    main()
