import socket

versionPacket  = "\x00\x5a\x00\x00\x01\x00\x00\x00"
versionPacket += "\x01\x36\x01\x2c\x00\x00\x08\x00"
versionPacket += "\x7f\xff\x7f\x08\x00\x00\x00\x01"
versionPacket += "\x00\x20\x00\x3a\x00\x00\x00\x00"
versionPacket += "\x00\x00\x00\x00\x00\x00\x00\x00"
versionPacket += "\x00\x00\x00\x00\x34\xe6\x00\x00"
versionPacket += "\x00\x01\x00\x00\x00\x00\x00\x00"
versionPacket += "\x00\x00"
versionPacket += "(CONNECT_DATA=(COMMAND=version))"

class TNSPacket:

    version = 10

    # :1 is the size + 58 of the packet
    basePacket  = "\x00:1\x00\x00\x01\x00\x00\x00"
    basePacket += "\x01\x36\x01\x2c\x00\x00\x08\x00"
    basePacket += "\x7f\xff\x7f\x08\x00\x00\x00\x01"
    # :2 is the real size of the packet
    basePacket += "\x00:2\x00\x3a\x00\x00\x00\x00"
    basePacket += "\x00\x00\x00\x00\x00\x00\x00\x00"
    basePacket += "\x00\x00\x00\x00\x34\xe6\x00\x00"
    basePacket += "\x00\x01\x00\x00\x00\x00\x00\x00"
    basePacket += "\x00\x00"

    # :1 is the size + 58 of the packet
    base10gPacket  = "\x00:1\x00\x00\x01\x00\x00\x00"
    base10gPacket += "\x01\x39\x01\x2c\x00\x81\x08\x00"
    base10gPacket += "\x7f\xff\x7f\x08\x00\x00\x01\x00"
    # :2 is the real size of the packet
    base10gPacket += "\x00:2\x00\x3a\x00\x00\x07\xf8"
    base10gPacket += "\x0c\x0c\x00\x00\x00\x00\x00\x00"
    base10gPacket += "\x00\x00\x00\x00\x00\x00\x00\x00"
    base10gPacket += "\x00\x00\x00\x00\x00\x00\x00\x00"
    base10gPacket += "\x00\x00"

    def getPacket(self, cmd):
        hLen1 = len(cmd) + 58
        hLen2 = len(cmd)
        hLen1 = eval("'\\" + str(hex(hLen1))[1:] + "'")
        hLen2 = eval("'\\" + str(hex(hLen2))[1:] + "'")

        if self.version >= 10:
            data = self.base10gPacket
        else:
            data = self.basePacket

        data = data.replace(":1", hLen1)
        data = data.replace(":2", hLen2)
        data += cmd

        return data

class TNS:

    TNS_TYPE_ACCEPT = 0
    TNS_TYPE_ERROR  = 1

    TNS_V7  =  7
    TNS_V8  =  8
    TNS_V9  =  9
    TNS_V10 = 10
    TNS_V11 = 11
    
    tns_data = None
    banner = None

    def sendCommand(self, command):
        pass

    def sendData(self, data):
        pass

    def sendConnectRequest(self, mSocket, mBuf):
        mSocket.send(mBuf)

    def recvTNSPkt(self, mSocket):
        packet = ""
        packet = mSocket.recv(1024)

        self.tns_data = packet

        if packet.find("(ERR=0)") > 0:
            self.packet_type = TNS.TNS_TYPE_ACCEPT
        else:
            self.packet_type = TNS.TNS_TYPE_ERROR

    def recvAcceptData(self, mSocket, mData):
        packet = ""
        packet = mSocket.recv(1024)

        self.banner = packet

        return(self.tns_data + self.banner)

    def getVSNNUM(self, verInfo):
        pos    = verInfo.find("VSNNUM=")
        endPos = verInfo.find(")", pos)
        data = verInfo[pos+7:endPos]

        return data

    def assignVersion(self, data):

        if not str(data).isalnum():
            return

        v = hex(int(data))
        v = v[0:3]
        v = eval(v)

        return(v)

class TNSCONNECT:

    def getVersionCommand(self):
        global versionPacket

        return versionPacket

class TNSParser:

    data = ""

    def __init__(self, data):
        if data:
            self.data = data
    
    def getValueFor(self, thekey, single = False):
        buf = []
        level = 0
        value = False
        flag = False
        word = ""

        for char in self.data:
            if char == "(":
                level += 1
                key = ""
                if value:
                    flag = True

                word = ""
            elif char == ")":
            
                if key.lower() == thekey.lower():
                    if not single:
                        buf.append(word)
                    else:
                        single = word
                        break

                level -= 1
                value = False
                word = ""
            elif char == "=":
                value = True
                key = word
                word = ""
            else:
                word += char

        return buf

class TNSDataFormatter:

    data = ""

    def __init__(self, data):
        if data:
            self.data = data
    
    def format(self):
        buf = "\r\n"
        level = 0
        value = False
        flag = False
        word = ""

        for char in self.data:
            if char == "(":
                level += 1
                if value:
                    flag = True
                    buf += "\r\n"

                buf += "  " + "  "*level
                word = ""
            elif char == ")":
                level -= 1
                value = False
                buf += "\r\n"
                word = ""
            elif char == "=":
                value = True
                buf += ": "
                word = ""
            else:
                buf += char
                word += char

        return buf
